import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

/**
 * abstract super class of all tetrominos; contains the act-method
 * generate a dummy image object and additional 4 Block objects moved within the Tetromino object (be free if the Tetromino deleted) 
 * @author Dietrich Boles (modified by Ulf Erben)
 * @version 1.0 (30.10.2008)
 */
public abstract class Tetromino1 extends Actor {

	// possible directions of a tetromino
	protected static final int NORTH = 0;

	protected static final int WEST = 1;

	protected static final int SOUTH = 2;

	protected static final int EAST = 3;

	protected static final int START = 18;

	protected Block[] b; // each tetromino consists of four blocks

	int direction; // direction of the tetromino

	boolean dead; // is the tetromino dead?

	private int counter; // internal counter

	Tetromino1(String color) {
		setImage("empty_back.png");
		b = new Block[4];
		for (int i = 0; i < 4; i++) {
			b[i] = new Block(color);
		}
		counter = 0;
		dead = false;
		DuoTetrisWorld.getWorld().addObject(this, 0,DuoTetrisWorld.getWorld().getHeight() - 1);
	}

	// changes the direction of a tetromino; the current direction is stored in
	// attribute direction
	abstract protected void setDirection();

	// left most block of the tetromino (depending on its direction)
	abstract protected Block leftMost();

	// right most block of the tetromino (depending on its direction)
	abstract protected Block rightMost();

	// deletes the four blocks of a tetromino
	void delete() {
		for (int i = 0; i < 4; i++) {
			getWorld().removeObject(b[i]);
		}
		dead = true;
	}

	// the current tetromino (more precisely its blocks) are falling down
	public void act() {
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		if (world.getCurrentTetromino1() == null) { // game ended
		    world.gameOver();
			return;
		}

	    if (dead)
			return;
			
		// checking user interactions
	    boolean keyAction = false;
		if (Greenfoot.isKeyDown("j") && counter < 4) {
			if (left()) {
				counter++;
				keyAction = true;
			}
		}
		if (Greenfoot.isKeyDown("l") && counter < 4) {
			if (right()) {
				counter++;
				keyAction = true;
			}
		}
		if (Greenfoot.isKeyDown("i") && counter < 3) {
			if (turnLeft()) {
				counter++;
				keyAction = true;
			}
		}
		if (Greenfoot.isKeyDown("k") || Greenfoot.isKeyDown(" ")) {
			down();
			return;
		}

		if (keyAction) {
			return;
		}

		// one row down
		oneDown();
		counter = 0;
	}

	// one column to the left
	boolean left() {
		if (leftOccupied())
			return false;
		for (int i = 0; i < 4; i++) {
			b[i].setLocation(b[i].getX() - 1, b[i].getY());
		}
		return true;
	}

	// left shift possible?
	boolean leftOccupied() {
	    DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		if (leftMost().getX() == world.PLAYFIELDSTART1) {
			return true;
		}
		blocks: for (int i = 0; i < 4; i++) {
			java.util.List list = world.getObjectsAt(b[i].getX() - 1, b[i]
					.getY(), Block.class);
			if (list.size() == 0) {
				continue;
			}
			// then list.size() == 1
			Actor a = (Actor) list.get(0);
			for (int j = 0; j < 4; j++) {
				if (i == j)
					continue;
				if (a == b[j])
					continue blocks;
			}
			return true;
		}
		return false;
	}

	// one column to the right
	boolean right() {
		if (rightOccupied())
			return false;
		for (int i = 0; i < 4; i++) {
			b[i].setLocation(b[i].getX() + 1, b[i].getY());
		}
		return true;
	}

	// right shif possible?
	boolean rightOccupied() {
        DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		if (rightMost().getX() == world.PLAYFIELDEND1 - 1) {
			return true;
		}
		blocks: for (int i = 0; i < 4; i++) {
			java.util.List list = world.getObjectsAt(b[i].getX() + 1, b[i]
					.getY(), Block.class);
			if (list.size() == 0) {
				continue;
			}
			// then list.size() == 1
			Actor a = (Actor) list.get(0);
			for (int j = 0; j < 4; j++) {
				if (i == j)
					continue;
				if (a == b[j])
					continue blocks;
			}
			return true;
		}
		return false;
	}

	// change the direction of the tetromino
	boolean turnLeft() {
		int oldDir = direction;
		direction = (direction + 1) % 4;
		setDirection();
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		for (int i = 0; i < 4; i++) {
			java.util.List list = world.getObjectsAt(b[i].getX(), b[i].getY(),
					null);
			if (list.size() > 1) {
				direction = oldDir;
				setDirection();
				return false;
			}
		}
		return true;
	}

	// tetromino slides one row down
	boolean oneDown() {
		// checks whether the tetromino is on the bottom row
		for (int i = 0; i < 4; i++) {
			if (!blockFree(i)) {
				checkRows();
				die();
				return false;
			}
		}

		// falling down
		for (int i = 0; i < 4; i++) {
			b[i].setLocation(b[i].getX(), b[i].getY() + 1);
		}
		return true;
	}

	// the tetromino is sliding to the bottom row
	void down() {
		while (oneDown())
			;
	}

	// is the cell below the block free?
	boolean blockFree(int index) {
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		java.util.List list = world.getObjectsAt(b[index].getX(), b[index]
				.getY() + 1, null);
		if (list.size() == 0) {
			return true;
		}
		// then list.size() == 1
		Actor a = (Actor) list.get(0);
		for (int i = 0; i < 4; i++) {
			if (i == index)
				continue;
			if (a == b[i])
				return true;
		}
		return false;
	}

	// checks whether there exists completed rows which can be removed
	void checkRows() {
		int noOfRows = 0;
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		rows: for (int r = world.getHeight() - 3; r >= 0; r--) {
			cols: for (int c = world.PLAYFIELDSTART1; c < world.PLAYFIELDEND1; c++) {
				java.util.List blocks = world.getObjectsAt(c, r, Block.class);
				if (blocks.size() == 0) {
					continue rows; // next row
				}
			}
			// clear row
			clearRow(r);
			noOfRows++;
			landslide(r);
			r++;
		}
		if (noOfRows > 0) {
			world.newPoints1(noOfRows);
		}
	}

	// removes the blocks of a row
	void clearRow(int row) {
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		for (int c = world.PLAYFIELDSTART1; c < world.PLAYFIELDEND1; c++) {
			world.removeObjects(world.getObjectsAt(c, row, Block.class));
		}
	}

	// performs a "landslide"
	void landslide(int row) {
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		for (int r = row - 1; r >= 0; r--) {
			for (int c = world.PLAYFIELDSTART1; c < world.PLAYFIELDEND1; c++) {
				java.util.List blocks = world.getObjectsAt(c, r, Block.class);
				if (blocks.size() > 0) {
					Block block = (Block) blocks.get(0);
					block.setLocation(block.getX(), block.getY() + 1);
				}
			}
		}
	}

	// kills the tetromino
	void die() {
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		world.removeObject(this);
		dead = true;
		Tetromino1 tetro = world.genTetromino1();
		if (checkEnd(tetro)) {
			tetro.delete();
			world.setCurrentTetromino1(null);
			world.gameOverFlag1 = true;
			world.gameOver();
		} else {
			DuoTetrisWorld.getWorld().setCurrentTetromino1(tetro);
		}
	}

	// checks for the new Tetromino start place if there existing blocks - end of game
	boolean checkEnd(Tetromino1 tetro) {
		DuoTetrisWorld world = DuoTetrisWorld.getWorld();
		for (int i = 0; i < 4; i++) {
			java.util.List list = world.getObjectsAt(tetro.b[i].getX(),
					tetro.b[i].getY(), Block.class);
			if (list.size() > 1) {
				return true;
			}
		}
		return false;
	}


	// return the number of digits of a number
	int length(int number) {
		if (number < 10)
			return 1;
		return length(number / 10) + 1;
	}
}
